// SPDX-License-Identifier: GPL-2.0-or-later

/***************************************************************************
 *   Copyright (C) 2019 Icenowy Zheng <icenowy@aosc.io>                    *
 *   Copyright (C) 2019 Caleb Szalacinski <contact@skiboy.net>             *
 *   Copyright (C) 2023 Benjamin VERNOUX 04 Sep 2023 (based on swm050.c)   *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "imp.h"
#include <helper/time_support.h>
#include <target/image.h>

#define CX32L003_FLASH_INIT_TIMEOUT 500 // Delay in ms
#define CX32L003_FLASH_ERASE_TIMEOUT 10000 // Delay in ms
#define CX32L003_FLASH_PROGRAM_TIMEOUT 5000 // Delay in ms
#define CX32L003_IWDT_RELOAD_TIMEOUT 10000 // Delay in ms

/* CX32L003F8 64KB Flash (128 sectors * 512 bytes) */
#define CX32L003_FLASH_PAGE_SIZE	512
#define CX32L003_FLASH_PAGES		128

/* Flash Erase/Program Algorithm */
#define FLASH_SIZE_32K 0x8000U
#define FLASH_SIZE_64K 0x10000U

// CX32L003 memory-mapped registers

#define SYS_HIRC_24M_ADDR	0x180000A0UL

#define RCC_UNLOCK_ADDR		0x40020060UL
#define RCC_HIRCCR_ADDR		0x40020028UL
#define RCC_SYSCLKCR_ADDR	0x40020020UL

#define RCC_HCLKEN_ADDR		0x40020008UL

#define FLASH_CR_ADDR		0x40020400UL
#define FLASH_BYPASS_ADDR	0x4002040CUL
#define FLASH_SLOCK0_ADDR	0x40020410UL
#define FLASH_SLOCK1_ADDR	0x40020414UL

#define IWDG_CFGR_ADDR		0x40002404UL
#define IWDG_RLOAD_ADDR		0x40002408UL
#define IWDG_UNLOCK_ADDR	0x40002418UL

#define DUMMY_DATA			0x5555AAAAUL

static void cx32l003_iwdt_reload(struct target* target);

static int cx32l003_init(struct target* target)
{
	long long start_ms;
	long long elapsed_ms;
	uint32_t val;
	int retval;

	/* RCC_UNLOCK UNLOCK */
	target_write_u32(target, RCC_UNLOCK_ADDR, 0x55AA6699);

	/* If IWDG is running, reload it to max value about 27s */
	cx32l003_iwdt_reload(target);

	/* Flash controller block clock enable */
	target_read_u32(target, RCC_HCLKEN_ADDR, &val);
	val = val | 0x100; // FlashCKE bit 8
	target_write_u32(target, IWDG_UNLOCK_ADDR, val);

	/* Wait for end of flash init "idle state" */
	start_ms = timeval_ms();
	do {
		retval = target_read_u32(target, FLASH_CR_ADDR, &val);
		if (retval != ERROR_OK) {
			LOG_ERROR("cx32l003_init: target_read_u32() fail");
			retval = ERROR_FAIL;
			break;
		}
		elapsed_ms = timeval_ms() - start_ms;
		if (elapsed_ms > 500)
			keep_alive();
		if (elapsed_ms > CX32L003_FLASH_INIT_TIMEOUT) {
			LOG_ERROR("cx32l003_init: flash init timeout");
			retval = ERROR_TIMEOUT_REACHED;
			break;
		}
	} while ((val & 4) == 4);

	/* If IWDG is running, reload it to max value about 27s */
	cx32l003_iwdt_reload(target);

	return retval;
}

static void cx32l003_iwdt_reload(struct target* target)
{
	uint32_t val;
	/* If IWDG is running, reload it to max value about 27s */
	target_read_u32(target, IWDG_CFGR_ADDR, &val);
	if ((val & 4) == 4) {
		target_write_u32(target, IWDG_UNLOCK_ADDR, 0x55aa6699);
		target_write_u32(target, IWDG_RLOAD_ADDR, 0xfffff);
		target_write_u32(target, IWDG_UNLOCK_ADDR, 0x55aa6698);
	}
}

/* Unlock All the FLASH Page Erase/write access. */
static void cx32l003_flash_unlock_all(struct target* target)
{
	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_SLOCK0_ADDR, 0xFFFFFFFF);

	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_SLOCK1_ADDR, 0xFFFFFFFF);
}
/* Lock All the FLASH Page Erase/write access. */
static void cx32l003_flash_lock_all(struct target* target)
{
	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_SLOCK0_ADDR, 0);

	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_SLOCK1_ADDR, 0);
}

static int cx32l003_flash_erase_sector(struct target* target, uint32_t flash_addr_dst)
{
	long long start_ms;
	long long elapsed_ms;
	uint32_t val;
	int retval;

	cx32l003_init(target);

	if (flash_addr_dst < FLASH_SIZE_32K) {
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_read_u32(target, FLASH_SLOCK0_ADDR, &val);
		val = val | (1 << (flash_addr_dst >> 10));
		target_write_u32(target, FLASH_SLOCK0_ADDR, val);
	} else if (flash_addr_dst < FLASH_SIZE_64K) {
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_read_u32(target, FLASH_SLOCK1_ADDR, &val);
		val = val | (1 << ((flash_addr_dst - FLASH_SIZE_32K) >> 10));
		target_write_u32(target, FLASH_SLOCK1_ADDR, val);
	}
	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_CR_ADDR, 2); // Flash Sector Erase
	/* Erase flash sector */
	target_write_u32(target, flash_addr_dst, DUMMY_DATA);
	/* After flash erase sector command wait at least 10 ms
	to avoid an error on target_read_u32() as bus is staled */
	alive_sleep(10);

	/* Wait for end of erase "idle state" */
	start_ms = timeval_ms();
	do {
		retval = target_read_u32(target, FLASH_CR_ADDR, &val);
		if (retval != ERROR_OK) {
			LOG_ERROR("cx32l003_flash_erase_sector: target_read_u32() fail");
			retval = ERROR_FAIL;
			break;
		}
		elapsed_ms = timeval_ms() - start_ms;
		if (elapsed_ms > 500)
			keep_alive();
		if (elapsed_ms > CX32L003_FLASH_ERASE_TIMEOUT) {
			LOG_ERROR("cx32l003_flash_erase_sector: flash erase timeout");
			retval = ERROR_TIMEOUT_REACHED;
			break;
		}
	} while ((val & 4) == 4);

	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_read_u32(target, FLASH_CR_ADDR, &val);
	target_write_u32(target, FLASH_CR_ADDR, val & 0xffffff00);

	if (flash_addr_dst < FLASH_SIZE_32K) {
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_read_u32(target, FLASH_SLOCK0_ADDR, &val);
		val = val & ~(1 << (flash_addr_dst >> 10));
		target_write_u32(target, FLASH_SLOCK0_ADDR, val);
	} else if (flash_addr_dst < FLASH_SIZE_64K) {
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_read_u32(target, FLASH_SLOCK1_ADDR, &val);
		val = val & ~(1 << ((flash_addr_dst - FLASH_SIZE_32K) >> 10));
		target_write_u32(target, FLASH_SLOCK1_ADDR, val);
	}

	return retval;
}

static int cx32l003_flash_mass_erase(struct target* target)
{
	long long start_ms;
	long long elapsed_ms;
	uint32_t val;
	int retval;

	cx32l003_init(target);

	/* Unlock all flash sectors for mass erase */
	cx32l003_flash_unlock_all(target);

	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_write_u32(target, FLASH_CR_ADDR, 3); // Flash Chip Erase
	/* Erase all flash sectors */
	target_write_u32(target, 0, DUMMY_DATA);
	/* After erase flash command  wait at least 250 ms
	to avoid an error on target_read_u32() as bus is staled */
	alive_sleep(250);

	/* Wait for end of erase "idle state" */
	start_ms = timeval_ms();
	do {
		retval = target_read_u32(target, FLASH_CR_ADDR, &val);
		if (retval != ERROR_OK) {
			LOG_ERROR("cx32l003_flash_erase_sector: target_read_u32() fail");
			retval = ERROR_FAIL;
			break;
		}
		elapsed_ms = timeval_ms() - start_ms;
		if (elapsed_ms > 500)
			keep_alive();
		if (elapsed_ms > CX32L003_FLASH_ERASE_TIMEOUT) {
			LOG_ERROR("cx32l003_flash_erase_sector: flash erase timeout");
			retval = ERROR_TIMEOUT_REACHED;
			break;
		}
	} while ((val & 4) == 4);

	target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
	target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
	target_read_u32(target, FLASH_CR_ADDR, &val);
	target_write_u32(target, FLASH_CR_ADDR, val & 0xffffff00);

	/* Lock all the flash sectors for erase/program */
	cx32l003_flash_lock_all(target);

	return retval;
}

int cx32l003_flash_program(struct target *target, uint32_t flash_addr_dst, int size, const uint32_t* ram_data_src)
{
	long long iwdt_start_ms;
	long long start_ms;
	long long elapsed_ms;
	uint32_t val;
	int retval = ERROR_OK;

	cx32l003_init(target);
	iwdt_start_ms = timeval_ms();

	/* Unlock all the flash sectors for program */
	cx32l003_flash_unlock_all(target);

	/* Program each word (32bits) */
	for (; size != 0; size = size - 4) {
		target_read_u32(target, FLASH_CR_ADDR, &val);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_write_u32(target, FLASH_CR_ADDR, val & 0xffffff00);
		target_read_u32(target, FLASH_CR_ADDR, &val);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0x5a5a);
		target_write_u32(target, FLASH_BYPASS_ADDR, 0xa5a5);
		target_write_u32(target, FLASH_CR_ADDR, val | 1); // Flash Program
		/* Program flash */
		target_write_u32(target, flash_addr_dst, *ram_data_src);
		/* After program flash command  wait at least 10 us
		to avoid an error on target_read_u32() as bus is staled */
		usleep(10);

		/* Wait end of program "idle state" */
		start_ms = timeval_ms();
		do {
			target_read_u32(target, FLASH_CR_ADDR, &val);
			if (retval != ERROR_OK) {
				LOG_ERROR("cx32l003_flash_program: target_read_u32() fail");
				retval = ERROR_FAIL;
				break;
			}
			elapsed_ms = timeval_ms() - start_ms;
			if (elapsed_ms > 500)
				keep_alive();
			if (elapsed_ms > CX32L003_FLASH_PROGRAM_TIMEOUT) {
				LOG_ERROR("cx32l003_flash_program: flash program timeout");
				retval = ERROR_TIMEOUT_REACHED;
				break;
			}
		} while ((val & 4) == 4);

		/* Check if IWDT shall be refreshed */
		elapsed_ms = timeval_ms() - iwdt_start_ms;
		if(elapsed_ms > CX32L003_IWDT_RELOAD_TIMEOUT) {
			LOG_INFO("cx32l003_flash_program: iwdt_start_ms elapsed_ms=%d => call cx32l003_iwdt_reload()", (int)elapsed_ms);
			cx32l003_iwdt_reload(target);
			iwdt_start_ms = timeval_ms();
		}
		flash_addr_dst = flash_addr_dst + 4;
		ram_data_src++;
	}

	/* Lock all the flash sectors for erase/program */
	cx32l003_flash_lock_all(target);

	return retval;
}

static int cx32l003_erase(struct flash_bank *bank, unsigned int first, unsigned int last)
{
	struct target *target = bank->target;
	int retval;
	long long start_ms;
	int elapsed_ms;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}
	LOG_INFO("cx32l003_erase first=%d last=%d", first, last);
	start_ms = timeval_ms();
	for (unsigned int curr_page = first; curr_page <= last; curr_page++) {
		uint32_t curr_addr = bank->base + (CX32L003_FLASH_PAGE_SIZE * curr_page);
		/* Perform sector erase */
		retval = cx32l003_flash_erase_sector(target, curr_addr);
		if (retval != ERROR_OK)
			return retval;
	}
	elapsed_ms = (int)(timeval_ms() - start_ms);
	LOG_INFO("cx32l003_erase retval=%d elapsed_ms=%d", retval, elapsed_ms);

	return ERROR_OK;
}

static int cx32l003_write(struct flash_bank *bank, const uint8_t *buffer, uint32_t offset, uint32_t count)
{
	struct target *target = bank->target;
	int retval;
	long long start_ms;
	int elapsed_ms;
	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		retval = ERROR_TARGET_NOT_HALTED;
		return retval;
	}

	/* Perform write */
	LOG_INFO("cx32l003_write buffer=%p offset=0x%08X count=%d", buffer, offset, count);
	start_ms = timeval_ms();
	retval = cx32l003_flash_program(target, offset, count, (const uint32_t *)buffer);
	elapsed_ms = (int)(timeval_ms() - start_ms);
	LOG_INFO("cx32l003_write retval=%d elapsed_ms=%d", retval, elapsed_ms);

	return retval;
}

static int cx32l003_probe(struct flash_bank *bank)
{
	return ERROR_OK;
}

static int cx32l003_mass_erase(struct flash_bank *bank)
{
	struct target *target = bank->target;
	int retval;
	long long start_ms;
	int elapsed_ms;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	/* Perform mass erase */
	LOG_INFO("cx32l003_mass_erase");
	start_ms = timeval_ms();
	retval = cx32l003_flash_mass_erase(target);
	elapsed_ms = (int)(timeval_ms() - start_ms);
	LOG_INFO("cx32l003_mass_erase retval=%d elapsed_ms=%d", retval, elapsed_ms);

	return retval;
}

COMMAND_HANDLER(cx32l003_handle_mass_erase_command)
{
	if (CMD_ARGC < 1)
		return ERROR_COMMAND_SYNTAX_ERROR;

	struct flash_bank *bank;
	int retval = CALL_COMMAND_HANDLER(flash_command_get_bank, 0, &bank);
	if (retval != ERROR_OK)
		return retval;

	retval = cx32l003_mass_erase(bank);
	if (retval == ERROR_OK)
		command_print(CMD, "cx32l003 mass erase complete");
	else
		command_print(CMD, "cx32l003 mass erase failed");

	return retval;
}

FLASH_BANK_COMMAND_HANDLER(cx32l003_flash_bank_command)
{
	free(bank->sectors);
	bank->write_start_alignment = 4;
	bank->write_end_alignment = 4;
	bank->size = CX32L003_FLASH_PAGE_SIZE * CX32L003_FLASH_PAGES;

	bank->num_sectors = CX32L003_FLASH_PAGES;
	bank->sectors = alloc_block_array(0, CX32L003_FLASH_PAGE_SIZE, CX32L003_FLASH_PAGES);
	if (!bank->sectors)
		return ERROR_FAIL;

	for (unsigned int i = 0; i < bank->num_sectors; i++)
		bank->sectors[i].is_protected = 0;

	return ERROR_OK;
}

static const struct command_registration cx32l003_exec_command_handlers[] = {
	{
		.name = "mass_erase",
		.handler = cx32l003_handle_mass_erase_command,
		.mode = COMMAND_EXEC,
		.usage = "bank_id",
		.help = "Erase entire flash device.",
	},
	COMMAND_REGISTRATION_DONE
};

static const struct command_registration cx32l003_command_handlers[] = {
	{
		.name = "cx32l003",
		.mode = COMMAND_ANY,
		.help = "cx32l003 flash command group",
		.usage = "",
		.chain = cx32l003_exec_command_handlers,
	},
	COMMAND_REGISTRATION_DONE
};

const struct flash_driver cx32l003_flash = {
	.name = "cx32l003",
	.commands = cx32l003_command_handlers,
	.flash_bank_command = cx32l003_flash_bank_command,
	.erase = cx32l003_erase,
	.write = cx32l003_write,
	.read = default_flash_read,
	.probe = cx32l003_probe,
	.auto_probe = cx32l003_probe,
	.erase_check = default_flash_blank_check,
	.free_driver_priv = default_flash_free_driver_priv,
};
